/**********************************************************************
$Id: mmclda.c,v 1.3 2002/10/14 09:10:08 bu Exp $

Copyright 2001-2002, Microbrew Software

Part of Microbrew Message Center
Multi-server Mailer Suite

http://www.microbrew.org/products/mmc/

by Bulent Yilmaz <bu@microbrew.org>
and James Brown <jim@microbrew.org>

Microbrew Message Center is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

Microbrew Message Center is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along with
Microbrew Message Center; if not, write to the Free Software Foundation,
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

**********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>

/* Configuration file from autoconf. */
#include "../config.h"

#if HAVE_FCNTL_H 
#include <fcntl.h>
#endif


#define BOXDIR PREFIX "/boxes"
#define VIRTUSERTABLE "/etc/mail/virtusertable.db"

#define LOGFILE PREFIX "/var/mmclda.log"
#ifndef TESTOUTPUT
#define TESTOUTPUT 1
#endif

int main(int argc, char* argv[]) {

    char *message=NULL;
    char *msgline;
    char *recline=NULL, *forplace=NULL;
    char *user, *domain;
    char *holder=NULL;
    int width=0;
    long int thisread=0, size=0, written=0;
    char mailbox[1024];
    int boxfd;
    char *deliveryuser=NULL;
    FILE *log=NULL;
    time_t now;
    struct stat filestat;
    char *cleanholder;

#if TESTOUTPUT    
    log=fopen(LOGFILE, "a");
#endif

    if(argc!=5) {
#if TESTOUTPUT
        if(log != NULL) fputs("Improper number of arguments.\n", log);
        fclose(log);
#endif
        /* Send proper DSN. */

        puts("5.5.4 Invalid command arguments");
        return 450;
    }
    
    msgline=malloc(65536);
    
    msgline=strncat(msgline, "From ", 66535);
    msgline=strncat(msgline, argv[2], 66535);
    msgline=strncat(msgline, " ", 66535);
    now=time(NULL);
    msgline=strncat(msgline, ctime(&now), 66535);
    
    size=strlen(msgline);
    
    message=strdup(msgline);
     
    user=argv[3];
    domain=argv[4];

    while((thisread = read(STDIN_FILENO, &msgline[0], 65536)) > 0 ) {

        size+=thisread;
        cleanholder=message;
        message=realloc(message, (size + 2) );

        if(message) {
            memcpy((message+size-thisread), msgline, thisread);
            *(message+size)=0;
        } else {
#if TESTOUTPUT
            if(log != NULL) fputs("realloc() failed.  Exiting.", log);
            fclose(log);
#endif      
            if(cleanholder != NULL) free(cleanholder);
            
            /* Clean up a few other variables. */
            
            free(message);
            
            /* Return proper DSN. */
            
            puts("4.6.5 Conversion Failed");
            return 450;
        }
    }
    
    *(message+size)='\n';
    *(message+size+1)='\0';
    
    size+=1;

    snprintf(mailbox, 1024, "%s/%s/mail/%s", BOXDIR, domain, user);
#if TESTOUTPUT
    if(log!=NULL) fprintf(log, "Mailbox = %s!\n", mailbox);
#endif
    
    if(!stat(mailbox, &filestat)) {
        /* Mailbox exists */
        if((filestat.st_size + size) > MAILBOXQUOTA) {
            puts("Mailbox exceeds quota.");
            
            free(message);
            
            /* Return proper DSN. */
            
            puts("4.2.2 Mailbox full");
            return 450;
        }
    } else {
        /* Mailbox doesn't exist.  Return error. */
#if TESTOUTPUT
        if(log!=NULL) fputs("Mailbox does not exist.", log);
        fclose(log);
#endif        
        
        free(message);
        
        /* Return proper DSN. */
        
        puts("5.1.1 Bad destination mailbox address");
        return 450;
    }
    
    if((boxfd=open(mailbox, O_WRONLY|O_APPEND)) > 0) {
        while(!written) {
            /* Make sure we're only writing to a locked file.  */
            
#if HAVE_FLOCK
            if(!flock(boxfd, LOCK_EX)) {
#else
#if HAVE_LOCKF
            if(!lockf(boxfd, F_LOCK, 0)) {
#endif
#endif
                while((written < size)&&(written>=0)) {
                    written+=write(boxfd, (message+written), (size-written));
                }
#if HAVE_FLOCK
                flock(boxfd, LOCK_UN);
#else
#if HAVE_LOCKF
                lockf(boxfd, F_ULOCK, 0);
#endif
#endif                
                if(written < 0) {
#if TESTOUTPUT                
                    if(log!=NULL) fputs("Could not write message to disk.", log);
                    fclose(log);
#endif              
                    free(message);      
                    
                    /* Return proper DSN. */
                    
                    puts("4.3.0 Undefined mail system status");
                    return 450;
                }
            } 
#if TESTOUTPUT
            else if(log!=NULL) fputs("Could not get lock.  Trying again.", log);
#endif
        }
        close(boxfd);
    } else {
#if TESTOUTPUT
        if(log!=NULL) fputs("Could not open mailbox.", log);
        fclose(log);
#endif
        
        /* Return proper DSN. */
        
        puts("4.2.1 Mailbox not accepting messages");
        return 450;
    }
#if TESTOUTPUT    
    if(log!=NULL) fprintf(log, "Message:\n%s\n", message);
    fclose(log);
#endif
    
    /* Return proper DSN. */
    
    puts("2.1.5 Successful delivery");
    
    free(message);
    
    return 0;
}
